@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003 - 2018 GraphicsMagick Group
+% Copyright (C) 2003 - 2020 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -195,9 +195,9 @@ static unsigned int IsPCX(const unsigned char *magick,const size_t length)
 */
 #define ThrowPCXReaderException(code_,reason_,image_) \
 { \
-  MagickFreeMemory(page_table) \
-  MagickFreeMemory(pcx_pixels); \
-  MagickFreeMemory(scanline); \
+  MagickFreeResourceLimitedMemory(page_table) \
+  MagickFreeResourceLimitedMemory(pcx_pixels); \
+  MagickFreeResourceLimitedMemory(scanline); \
   ThrowReaderException(code_,reason_,image_); \
 }
 static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
@@ -282,7 +282,7 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
       magic=ReadBlobLSBLong(image);
       if (magic != 987654321)
         ThrowPCXReaderException(CorruptImageError,ImproperImageHeader,image);
-      page_table=MagickAllocateArray(ExtendedSignedIntegralType *,
+      page_table=MagickAllocateResourceLimitedArray(ExtendedSignedIntegralType *,
                                      1024,sizeof(ExtendedSignedIntegralType));
       if (page_table == (ExtendedSignedIntegralType *) NULL)
         ThrowPCXReaderException(ResourceLimitError,MemoryAllocationFailed,image);
@@ -356,7 +356,7 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
                             "    left=%u\n"
                             "    top=%u\n"
                             "    right=%u\n"
-                          "    bottom=%u\n"
+                            "    bottom=%u\n"
                             "    horizontal_resolution=%u\n"
                             "    vertical_resolution=%u\n"
                             "    reserved=%u\n"
@@ -435,7 +435,7 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
     if (pcx_info.planes > 6)
       ThrowPCXReaderException(CorruptImageError,UnsupportedNumberOfPlanes,image);
 
-    if ((pcx_info.bits_per_pixel >= 8) || (pcx_info.planes != 1))
+   if ((pcx_info.bits_per_pixel >= 8) && (pcx_info.planes != 1))
       {
         image->storage_class=DirectClass;
         if (image->logging)
@@ -520,16 +520,16 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
                                 MagickArraySize(pcx_info.bytes_per_line,
                                                 pcx_info.planes));
     if ((0 == pcx_packets) ||
-        ((size_t) (pcx_info.bits_per_pixel*pcx_info.planes*image->columns) >
-         (pcx_packets*8U)))
+        (((size_t) pcx_info.bits_per_pixel*pcx_info.planes*image->columns) >
+         ((size_t) pcx_packets*8U)))
       ThrowPCXReaderException(CorruptImageError,ImproperImageHeader,image);
-    pcx_pixels=MagickAllocateMemory(unsigned char *,pcx_packets);
+    pcx_pixels=MagickAllocateResourceLimitedMemory(unsigned char *,pcx_packets);
     if (pcx_pixels == (unsigned char *) NULL)
       ThrowPCXReaderException(ResourceLimitError,MemoryAllocationFailed,image);
     scanline_size=MagickArraySize(Max(image->columns,
                                       (size_t) pcx_info.bytes_per_line),
                                   Max(pcx_info.planes,8));
-    scanline=MagickAllocateMemory(unsigned char *,scanline_size);
+    scanline=MagickAllocateResourceLimitedMemory(unsigned char *,scanline_size);
     if (scanline == (unsigned char *) NULL)
       ThrowPCXReaderException(ResourceLimitError,MemoryAllocationFailed,image);
     (void) memset(scanline,0,scanline_size);
@@ -604,7 +604,7 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
                   256 color images have their color map at the end of the file.
                 */
                 pcx_info.colormap_signature=ReadBlobByte(image);
-                (void) ReadBlob(image,3*image->colors,(char *) pcx_colormap);
+                (void) ReadBlob(image, (size_t) 3*image->colors,(char *) pcx_colormap);
                 p=pcx_colormap;
                 for (i=0; i < image->colors; i++)
                 {
@@ -620,7 +620,7 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
     */
     for (y=0; y < (long) image->rows; y++)
     {
-      p=pcx_pixels+((unsigned long) y*pcx_info.bytes_per_line*pcx_info.planes);
+      p=pcx_pixels+((size_t) y*pcx_info.bytes_per_line*pcx_info.planes);
       q=SetImagePixels(image,0,y,image->columns,1);
       if (q == (PixelPacket *) NULL)
         break;
@@ -773,14 +773,15 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
                                       image->columns,image->rows))
             break;
     }
-    MagickFreeMemory(scanline);
-    MagickFreeMemory(pcx_pixels);
+    MagickFreeResourceLimitedMemory(scanline);
+    MagickFreeResourceLimitedMemory(pcx_pixels);
     if (EOFBlob(image))
       {
         ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,
           image->filename);
         break;
       }
+    StopTimer(&image->timer);
     /*
       Proceed to next image.
     */
@@ -813,7 +814,7 @@ static Image *ReadPCXImage(const ImageInfo *image_info,ExceptionInfo *exception)
       }
   }
   if (page_table != (ExtendedSignedIntegralType *) NULL)
-    MagickFreeMemory(page_table);
+    MagickFreeResourceLimitedMemory(page_table);
   while (image->previous != (Image *) NULL)
     image=image->previous;
   CloseBlob(image);
@@ -976,9 +977,9 @@ static MagickPassFail WriteRLEPixels(Image *image,
 
 #define LiberatePCXAllocations()                \
   {                                             \
-    MagickFreeMemory(pcx_colormap);             \
-    MagickFreeMemory(pcx_pixels);               \
-    MagickFreeMemory(page_table);               \
+    MagickFreeResourceLimitedMemory(pcx_colormap);             \
+    MagickFreeResourceLimitedMemory(pcx_pixels);               \
+    MagickFreeResourceLimitedMemory(page_table);               \
   }
 
 #define ThrowPCXWriterException(code_,reason_,image_) \
@@ -1008,7 +1009,7 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
     *q;
 
   size_t
-    length;
+    bytes_per_line;
 
   unsigned char
     *pcx_colormap = (unsigned char *) NULL,
@@ -1056,7 +1057,7 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
       */
       write_dcx=MagickTrue;
       (void) WriteBlobLSBLong(image,0x3ADE68B1L);
-      page_table=MagickAllocateMemory(ExtendedSignedIntegralType *,
+      page_table=MagickAllocateResourceLimitedMemory(ExtendedSignedIntegralType *,
         1024*sizeof(ExtendedSignedIntegralType));
       if (page_table == (ExtendedSignedIntegralType *) NULL)
         ThrowPCXWriterException(ResourceLimitError,MemoryAllocationFailed,image);
@@ -1126,8 +1127,16 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
         if (image->matte)
           pcx_info.planes++;
       }
-    pcx_info.bytes_per_line=(unsigned short)
-      (((unsigned long) image->columns*pcx_info.bits_per_pixel+7)/8);
+
+    /* image->columns*pcx_info.bits_per_pixel+7)/8 */
+    bytes_per_line=MagickArraySize(image->columns,pcx_info.bits_per_pixel);
+    if (bytes_per_line && (~((size_t)0)-7 > bytes_per_line))
+      bytes_per_line += 7;
+    bytes_per_line /= 8;
+    pcx_info.bytes_per_line=(unsigned short) bytes_per_line;
+    if ((pcx_info.bytes_per_line == 0) ||
+        ((size_t) pcx_info.bytes_per_line != bytes_per_line))
+      ThrowPCXWriterException(CoderError,UnsupportedNumberOfColumns,image);
     pcx_info.palette_info=1;
     pcx_info.colormap_signature=0x0c;
     /*
@@ -1147,7 +1156,7 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
     /*
       Dump colormap to file.
     */
-    pcx_colormap=MagickAllocateMemory(unsigned char *,3*256);
+    pcx_colormap=MagickAllocateResourceLimitedMemory(unsigned char *,3*256);
     if (pcx_colormap == (unsigned char *) NULL)
       ThrowPCXWriterException(ResourceLimitError,MemoryAllocationFailed,image);
     for (i=0; i < (3*256); i++)
@@ -1168,8 +1177,7 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
     for (i=0; i < 58; i++)
       (void) WriteBlobByte(image,'\0');
     /* Allocate memory for one pixel row. */
-    length=(size_t) pcx_info.bytes_per_line*pcx_info.planes;
-    pcx_pixels=MagickAllocateMemory(unsigned char *,length);
+    pcx_pixels=MagickAllocateResourceLimitedArray(unsigned char *,bytes_per_line,pcx_info.planes);
     if (pcx_pixels == (unsigned char *) NULL)
       ThrowPCXWriterException(ResourceLimitError,MemoryAllocationFailed,image);
     q=pcx_pixels;
@@ -1316,8 +1324,8 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
 
     (void) WriteBlobByte(image,pcx_info.colormap_signature);
     (void) WriteBlob(image,3*256,(char *) pcx_colormap);
-    MagickFreeMemory(pcx_pixels);
-    MagickFreeMemory(pcx_colormap);
+    MagickFreeResourceLimitedMemory(pcx_pixels);
+    MagickFreeResourceLimitedMemory(pcx_colormap);
     if (image->next == (Image *) NULL)
       break;
     image=SyncNextImageInList(image);
@@ -1342,7 +1350,7 @@ static unsigned int WritePCXImage(const ImageInfo *image_info,Image *image)
       (void) WriteBlobLSBLong(image,0x3ADE68B1L);
       for (i=0; i <= (long) scene; i++)
         (void) WriteBlobLSBLong(image,(unsigned long) page_table[i]);
-      MagickFreeMemory(page_table);
+      MagickFreeResourceLimitedMemory(page_table);
     }
   if (status == False)
     ThrowPCXWriterException(FileOpenError,UnableToWriteFile,image);
