@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2019 GraphicsMagick Group
+% Copyright (C) 2003-2022 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -179,7 +179,7 @@ static unsigned int PNMInteger(Image *image,const unsigned int base)
               }
           }
         length=MaxTextExtent;
-        comment=MagickAllocateMemory(char *,length+sizeof(P7Comment));
+        comment=MagickAllocateResourceLimitedMemory(char *,length+sizeof(P7Comment));
         p=comment;
         offset=p-comment;
         if (comment != (char *) NULL)
@@ -190,12 +190,18 @@ static unsigned int PNMInteger(Image *image,const unsigned int base)
                 size_t
                   text_length;
 
+                char
+                  *new_comment;
+
                 text_length=(size_t) (p-comment);
-                length<<=1;
                 length+=MaxTextExtent;
-                MagickReallocMemory(char *,comment,length+sizeof(P7Comment));
-                if (comment == (char *) NULL)
-                  break;
+                new_comment=MagickReallocateResourceLimitedMemory(char *,comment,length+sizeof(P7Comment));
+                if (new_comment == (char *) NULL)
+                  {
+                    MagickFreeResourceLimitedMemory(comment);
+                    break;
+                  }
+                comment=new_comment;
                 p=comment+text_length;
               }
             c=ReadBlobByte(image);
@@ -212,7 +218,7 @@ static unsigned int PNMInteger(Image *image,const unsigned int base)
           can span multiple lines.
         */
         (void) SetImageAttribute(image,"comment",comment);
-        MagickFreeMemory(comment);
+        MagickFreeResourceLimitedMemory(comment);
         continue;
       }
   } while (!isdigit(c));
@@ -588,7 +594,8 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
         ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,
                        image->filename);
 
-      if ((1 == samples_per_pixel) && (max_value < MaxColormapSize))
+      if ((1 == samples_per_pixel) && (max_value < MaxColormapSize) &&
+          ((size_t) image->columns*image->rows >= max_value))
         {
           image->storage_class=PseudoClass;
           image->colors=
@@ -961,7 +968,7 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
               }
             else
               {
-                bytes_per_row=MagickArraySize(((bits_per_sample+7)/8)*
+                bytes_per_row=MagickArraySize((((size_t) bits_per_sample+7)/8)*
                                               samples_per_pixel,image->columns);
               }
 
@@ -1146,14 +1153,18 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
             DestroyThreadViewDataSet(scanline_set);
             image->is_monochrome=is_monochrome;
             image->is_grayscale=is_grayscale;
+            if ((status == MagickFail) && (image->exception.severity))
+              CopyException(exception,&image->exception);
             if (EOFBlob(image))
-              ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,
-                             image->filename);
+              ThrowReaderException(CorruptImageError,UnexpectedEndOfFile,image);
             break;
           }
         default:
           ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
         }
+      StopTimer(&image->timer);
+      if (status ==MagickFail)
+        break;
       /*
         Proceed to next image.
       */
@@ -1653,7 +1664,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 
             value=(depth <=8 ? 255U : depth <= 16 ? 65535U : 4294967295U);
 
-            j += sprintf(&buffer[j],"%u\n",value);
+            j += snprintf(&buffer[j],(sizeof(buffer)-j),"%u\n",value);
             for (y=0; y < image->rows; y++)
               {
                 p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
@@ -1678,12 +1689,12 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
                     else if (depth <= 16)
                       {
                         value=ScaleQuantumToShort(index);
-                        j += sprintf(&buffer[j]," %u",value);
+                        j += snprintf(&buffer[j],(sizeof(buffer)-j)," %u",value);
                       }
                     else
                       {
                         value=ScaleQuantumToLong(index);
-                        j += sprintf(&buffer[j]," %u",value);
+                        j += snprintf(&buffer[j],(sizeof(buffer)-j)," %u",value);
                       }
 
                     i++;
@@ -1743,7 +1754,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 
             value=(depth <=8 ? 255U : (depth <= 16 ? 65535U : 4294967295U));
 
-            j += sprintf(&buffer[j],"%u\n",value);
+            j += snprintf(&buffer[j],(sizeof(buffer)-j),"%u\n",value);
             for (y=0; y < image->rows; y++)
               {
                 p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
@@ -1769,14 +1780,14 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
                       }
                     else if (depth <= 16)
                       {
-                        j += sprintf(&buffer[j],"%u %u %u ",
+                        j += snprintf(&buffer[j],(sizeof(buffer)-j),"%u %u %u ",
                                      ScaleQuantumToShort(p->red),
                                      ScaleQuantumToShort(p->green),
                                      ScaleQuantumToShort(p->blue));
                       }
                     else
                       {
-                        j += sprintf(&buffer[j],"%u %u %u ",
+                        j += snprintf(&buffer[j],(sizeof(buffer)-j),"%u %u %u ",
                                      (unsigned int) ScaleQuantumToLong(p->red),
                                      (unsigned int) ScaleQuantumToLong(p->green),
                                      (unsigned int) ScaleQuantumToLong(p->blue));
@@ -1938,7 +1949,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
               }
             else
               {
-                bytes_per_row=MagickArraySize(((bits_per_sample+7)/8)*
+                bytes_per_row=MagickArraySize((((size_t) bits_per_sample+7)/8)*
                                               samples_per_pixel,image->columns);
               }
 
@@ -1953,7 +1964,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
             /*
               Allocate memory for pixels.
             */
-            pixels=MagickAllocateMemory(unsigned char *,bytes_per_row);
+            pixels=MagickAllocateResourceLimitedMemory(unsigned char *,bytes_per_row);
             if (pixels == (unsigned char *) NULL)
               ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,
                                    image);
@@ -2035,7 +2046,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
                                                 image->columns,image->rows))
                       break;
               }
-            MagickFreeMemory(pixels);
+            MagickFreeResourceLimitedMemory(pixels);
 
             break;
           }
@@ -2081,11 +2092,11 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
                 for (i=0; i < 2; i++)
                   for (j=0; j < 16; j++)
                     {
-                      red_map[i][j]=MagickAllocateMemory(unsigned short *,
+                      red_map[i][j]=MagickAllocateResourceLimitedMemory(unsigned short *,
                                                          256*sizeof(unsigned short));
-                      green_map[i][j]=MagickAllocateMemory(unsigned short *,
+                      green_map[i][j]=MagickAllocateResourceLimitedMemory(unsigned short *,
                                                            256*sizeof(unsigned short));
-                      blue_map[i][j]=MagickAllocateMemory(unsigned short *,
+                      blue_map[i][j]=MagickAllocateResourceLimitedMemory(unsigned short *,
                                                           256*sizeof(unsigned short));
                       if ((red_map[i][j] == (unsigned short *) NULL) ||
                           (green_map[i][j] == (unsigned short *) NULL) ||
@@ -2094,9 +2105,9 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
                           for (i=0; i < 2; i++)
                             for (j=0; j < 16; j++)
                               {
-                                MagickFreeMemory(green_map[i][j]);
-                                MagickFreeMemory(blue_map[i][j]);
-                                MagickFreeMemory(red_map[i][j]);
+                                MagickFreeResourceLimitedMemory(green_map[i][j]);
+                                MagickFreeResourceLimitedMemory(blue_map[i][j]);
+                                MagickFreeResourceLimitedMemory(red_map[i][j]);
                               }
                           ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,
                                                image);
@@ -2173,9 +2184,9 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
                 for (i=0; i < 2; i++)
                   for (j=0; j < 16; j++)
                     {
-                      MagickFreeMemory(green_map[i][j]);
-                      MagickFreeMemory(blue_map[i][j]);
-                      MagickFreeMemory(red_map[i][j]);
+                      MagickFreeResourceLimitedMemory(green_map[i][j]);
+                      MagickFreeResourceLimitedMemory(blue_map[i][j]);
+                      MagickFreeResourceLimitedMemory(red_map[i][j]);
                     }
                 break;
           }
