@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2018 GraphicsMagick Group
+% Copyright (C) 2003-2022 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -21,7 +21,7 @@
 %                                                                             %
 %                              Software Design                                %
 %                              Jaroslav Fojtik                                %
-%                              June 2000 - 2018                               %
+%                              June 2000 - 2022                               %
 %                         Rework for GraphicsMagick                           %
 %                              Bob Friesenhahn                                %
 %                               Feb-May 2003                                  %
@@ -57,7 +57,7 @@ typedef struct
    } RGB_Record;
 
 /* Default palette for WPG level 1 */
-const RGB_Record WPG1_Palette[256]={
+static const RGB_Record WPG1_Palette[256]={
 {  0,  0,  0},          {  0,  0,168},
 {  0,168,  0},          {  0,168,168},
 {168,  0,  0},          {168,  0,168},
@@ -235,7 +235,6 @@ static unsigned int IsWPG(const unsigned char *magick,const size_t length)
   return(False);
 }
 
-
 static int Rd_WP_DWORD(Image *image, unsigned long *d)
 {
   unsigned char b;
@@ -382,7 +381,7 @@ return RetVal;
   x++; \
   if((long) x>=ldblk) \
   { \
-    if(InsertRow(BImgBuff,y,image,bpp)==MagickFail) RetVal=-6; \
+    if(InsertRow(BImgBuff,y,image,bpp)==MagickFail) { RetVal=-6; goto unpack_wpg_raser_error; } \
     x=0; \
     y++; \
     if(y>=image->rows) break; \
@@ -392,17 +391,30 @@ return RetVal;
 
 /** Call this function to ensure that all data matrix is filled with something. This function
  * is used only to error recovery. */
-static void ZeroFillMissingData(unsigned char *BImgBuff,unsigned long x, unsigned long y, Image *image,
-                                int bpp, long ldblk)
+static MagickPassFail ZeroFillMissingData(unsigned char *BImgBuff,unsigned long x, unsigned long y, Image *image,
+                                          int bpp, long ldblk)
 {
-  while(y < image->rows)
+  MagickPassFail
+    status = MagickPass;
+
+  while(y<image->rows && image->exception.severity!=UndefinedException)
   {
-    if((long) x<ldblk) memset(BImgBuff+x, 0, ldblk-(long)x);
-    if (InsertRow(BImgBuff,y,image,bpp) == MagickFail)
-      break;
-    x = 0;
+    if((long) x<ldblk)
+    {
+      memset(BImgBuff+x, 0, (size_t)ldblk-(size_t)x);
+      if(x == 0)
+        x = ldblk;      /* Do not memset any more */
+      else
+        x = 0;          /* Next pass will need to clear whole row */
+    }
+    if(InsertRow(BImgBuff,y,image,bpp) == MagickFail)
+      {
+        status = MagickFail;
+        break;
+      }
     y++;
   }
+  return status;
 }
 
 
@@ -434,7 +446,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
   (void) LogMagickEvent(CoderEvent,GetMagickModule(),
                         "Raster allocation size: %ld byte%s",
                         ldblk, (ldblk > 1 ? "s" : ""));
-  BImgBuff=MagickAllocateMemory(unsigned char *,(size_t) ldblk);
+  BImgBuff=MagickAllocateResourceLimitedMemory(unsigned char *,(size_t) ldblk);
   if(BImgBuff==NULL) return(-2);
   (void) memset(BImgBuff,0,(size_t) ldblk);
 
@@ -444,7 +456,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
       if(i==EOF)
         {
           ZeroFillMissingData(BImgBuff,x,y,image,bpp,ldblk);
-          MagickFreeMemory(BImgBuff);
+          MagickFreeResourceLimitedMemory(BImgBuff);
           return(-5);
         }
       bbuf = i;
@@ -476,7 +488,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
           if(i==EOF)
           {
             ZeroFillMissingData(BImgBuff,x,y,image,bpp,ldblk);
-            MagickFreeMemory(BImgBuff);
+            MagickFreeResourceLimitedMemory(BImgBuff);
             return -7;
           }
           RunCount = i;
@@ -488,7 +500,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
                 y++;
                 ZeroFillMissingData(BImgBuff,x,y,image,bpp,ldblk);
               }
-            MagickFreeMemory(BImgBuff);
+            MagickFreeResourceLimitedMemory(BImgBuff);
             return(-3);
           }
           for(i=0; i<(int)RunCount; i++)
@@ -496,14 +508,12 @@ static int UnpackWPGRaster(Image *image,int bpp)
                         /* when x=0; y points to a new empty line. For y=0 zero line will be populated. */
               if(y>=image->rows)
                 {
-                  ZeroFillMissingData(BImgBuff,x,y,image,bpp,ldblk);
-                  MagickFreeMemory(BImgBuff);
+                  MagickFreeResourceLimitedMemory(BImgBuff);
                   return(-4);
                 }
               if(InsertRow(BImgBuff,y,image,bpp)==MagickFail)
                 {
-                  ZeroFillMissingData(BImgBuff,x,y,image,bpp,ldblk);
-                  MagickFreeMemory(BImgBuff);
+                  MagickFreeResourceLimitedMemory(BImgBuff);
                   return(-6);
                 }
               y++;
@@ -511,7 +521,8 @@ static int UnpackWPGRaster(Image *image,int bpp)
         }
       }
     }
-  MagickFreeMemory(BImgBuff);
+unpack_wpg_raser_error:;
+  MagickFreeResourceLimitedMemory(BImgBuff);
   return(RetVal);
 }
 
@@ -526,7 +537,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
   x++; \
   if((long) x >= ldblk) \
   { \
-    if(InsertRow(BImgBuff,(long) y,image,bpp)==MagickFail) RetVal=-6; \
+    if(InsertRow(BImgBuff,(long) y,image,bpp)==MagickFail) { RetVal=-6; goto unpack_wpg2_error; } \
     x=0; \
     y++; \
     XorMe = 0; \
@@ -539,8 +550,8 @@ static int UnpackWPGRaster(Image *image,int bpp)
 #define FreeUnpackWPG2RasterAllocs(BImgBuff,UpImgBuff) \
   do \
     { \
-      MagickFreeMemory(BImgBuff); \
-      MagickFreeMemory(UpImgBuff); \
+      MagickFreeResourceLimitedMemory(BImgBuff); \
+      MagickFreeResourceLimitedMemory(UpImgBuff); \
     } while(0);
 
 /* WPG2 raster reader. */
@@ -574,10 +585,10 @@ static int UnpackWPG2Raster(Image *image, int bpp)
   x=0;
   y=0;
   ldblk=(long) ((bpp*image->columns+7)/8);
-  BImgBuff=MagickAllocateMemory(unsigned char *,(size_t) ldblk);
+  BImgBuff=MagickAllocateResourceLimitedMemory(unsigned char *,(size_t) ldblk);
   if(BImgBuff==NULL)
     return(-2);
-  UpImgBuff=MagickAllocateMemory(unsigned char *,(size_t) ldblk);
+  UpImgBuff=MagickAllocateResourceLimitedMemory(unsigned char *,(size_t) ldblk);
   if(UpImgBuff==NULL)
   {
     FreeUnpackWPG2RasterAllocs(BImgBuff,UpImgBuff);
@@ -703,6 +714,7 @@ static int UnpackWPG2Raster(Image *image, int bpp)
             }
         }
     }
+unpack_wpg2_error:;
   FreeUnpackWPG2RasterAllocs(BImgBuff,UpImgBuff);
   return(RetVal);
 }
@@ -710,7 +722,7 @@ static int UnpackWPG2Raster(Image *image, int bpp)
 
 typedef float tCTM[3][3];
 
-static unsigned LoadWPG2Flags(Image *image,char Precision,float *Angle,tCTM *CTM)
+static unsigned LoadWPG2Flags(Image *image, char Precision, float *Angle, tCTM *CTM)
 {
 const unsigned char TPR=1,TRN=2,SKW=4,SCL=8,ROT=0x10,OID=0x20,LCK=0x80;
 long x;
@@ -727,14 +739,21 @@ unsigned Flags;
  if(Flags & OID)
         {
         if(Precision==0)
-          {/*x=*/ (void) ReadBlobLSBShort(image);}      /*ObjectID*/
+          {x = ReadBlobLSBShort(image);
+           if(x >= 0x8000)
+             {
+               Precision = 1;   /* Double precision Switched on. */
+             (void)ReadBlobLSBShort(image);
+             /* ObjectId = ((xW & 0x7FFF)<<16) | DenX; */
+             }
+          }      /*ObjectID*/
         else
-          {/*x=*/ (void) ReadBlobLSBLong(image);}       /*ObjectID (Double precision)*/
+          {/*x=*/ (void) ReadBlobLSBLong(image);}  /*ObjectID native (Double precision)*/
         }
  if(Flags & ROT)
         {
         x=ReadBlobLSBLong(image);       /*Rot Angle*/
-        if(Angle) *Angle=x/65536.0;
+        if(Angle) *Angle=(float) (x/65536.0);
         }
  if(Flags & (ROT|SCL))
         {
@@ -752,13 +771,10 @@ unsigned Flags;
         }
  if(Flags & TRN)
         {
-        x=ReadBlobLSBLong(image); DenX=ReadBlobLSBShort(image);  /*Tx*/
-        if(x>=0) (*CTM)[0][2] = (float)x+(float)DenX/0x10000;
-            else (*CTM)[0][2] = (float)x-(float)DenX/0x10000;
-        x=ReadBlobLSBLong(image); DenX=ReadBlobLSBShort(image);  /*Ty*/
-        (*CTM)[1][2]=(float)x + ((x>=0)?1:-1)*(float)DenX/0x10000;
-        if(x>=0) (*CTM)[1][2] = (float)x+(float)DenX/0x10000;
-            else (*CTM)[1][2] = (float)x-(float)DenX/0x10000;
+        DenX=ReadBlobLSBShort(image); x=ReadBlobLSBLong(image);	 /*Tx*/
+        (*CTM)[0][2] = (double)x + (float)DenX/0x10000;
+        DenX=ReadBlobLSBShort(image); x=ReadBlobLSBLong(image);  /*Ty*/
+        (*CTM)[1][2] = (double)x + (float)DenX/0x10000;        
         }
  if(Flags & TPR)
         {
@@ -772,14 +788,9 @@ unsigned Flags;
 
 
 static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
-                                ExtendedSignedIntegralType PS_Offset,long PS_Size,ExceptionInfo *exception)
+                                ExtendedSignedIntegralType PS_Offset,
+                                size_t PS_Size,ExceptionInfo *exception)
 {
-  char
-    postscript_file[MaxTextExtent];
-
-  FILE
-    *ps_file;
-
   ImageInfo
     *clone_info;
 
@@ -787,77 +798,170 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
     *image2;
 
   unsigned char
-    magick[2*MaxTextExtent];
+    header_magick[2*MaxTextExtent];
+
+  void
+    *ps_data,
+    *ps_data_alloc = (unsigned char *) NULL;
+
+  char
+    format[MaxTextExtent];
 
   size_t
-    magick_size;
+    header_magick_size;
 
+  magick_off_t
+    filesize;
 
-  if ((clone_info=CloneImageInfo(image_info)) == NULL)
-    return(image);
-  clone_info->blob=(void *) NULL;
-  clone_info->length=0;
+  if (image->logging)
+    (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                          "ExtractPostscript(): PS_Offset=%"MAGICK_OFF_F"d, PS_Size=%"MAGICK_SIZE_T_F"u",
+                          (magick_off_t) PS_Offset, (MAGICK_SIZE_T) PS_Size);
 
-  /* Obtain temporary file */
-  ps_file = AcquireTemporaryFileStream(postscript_file,BinaryFileIOMode);
-  if (!ps_file)
+  /*
+    Validate that claimed subordinate image data is contained in file size
+  */
+  filesize = GetBlobSize(image);
+  if ((PS_Offset > filesize) || ((size_t) (filesize - PS_Offset) < PS_Size))
     {
-      (void) LogMagickEvent(CoderEvent,GetMagickModule(),"Gannot create file stream for PS image");
-      goto FINISH;
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "ExtractPostscript(): Failed to seek to PS_Offset=%"MAGICK_OFF_F"d",
+                              (magick_off_t) PS_Offset);
+      ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
+      return image;
     }
 
-  /* Copy postscript to temporary file */
-  if(SeekBlob(image,PS_Offset,SEEK_SET) != PS_Offset) goto BAD_SEEK;
-  magick_size = ReadBlob(image, sizeof(magick), magick);
-
-  if(SeekBlob(image,PS_Offset,SEEK_SET) != PS_Offset)
-  {
-BAD_SEEK:
-    (void) fclose(ps_file);
-    ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
-    goto FINISH_UNL;
-  }
-
-  while(PS_Size-- > 0)
+  /*
+    Get subordinate file header magick and use it to identify file format
+  */
+  if (SeekBlob(image,PS_Offset,SEEK_SET) != PS_Offset)
     {
-      int c;
-      if ((c = ReadBlobByte(image)) == EOF)
-        {
-          (void) fclose(ps_file);
-          ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
-          goto FINISH_UNL;
-        }
-      (void) fputc(c,ps_file);
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "ExtractPostscript(): Failed to seek to PS_Offset=%"MAGICK_OFF_F"d",
+                              (magick_off_t) PS_Offset);
+      ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
+      return image;
     }
-  (void) fclose(ps_file);
-
-  /* Detect file format - Check magic.mgk configuration file. */
-  if (GetMagickFileFormat(magick,magick_size,clone_info->magick,
-          MaxTextExtent,exception) == MagickFail)
+  header_magick_size = ReadBlob(image, Min(sizeof(header_magick),PS_Size), header_magick);
+  format[0]='\0';
+  /*
+    MagickExport MagickPassFail
+    GetMagickFileFormat(const unsigned char *header, const size_t header_length,
+      char *format, const size_t format_length,
+      ExceptionInfo *exception)
+  */
+  if (GetMagickFileFormat(header_magick,header_magick_size,format,
+                          sizeof(format),exception) == MagickFail)
     {
       (void) LogMagickEvent(CoderEvent,GetMagickModule(),
                             "Failed to identify embedded file type!");
       ThrowException(exception,CorruptImageError,UnableToReadImageHeader,image->filename);
-      goto FINISH_UNL;
+      return image;
     }
 
-  if(!ApproveFormatForWPG(clone_info->magick))
+  /*
+    Verify if this is an allowed subordinate image format
+  */
+  if(!ApproveFormatForWPG(format))
   {
     (void) LogMagickEvent(CoderEvent, GetMagickModule(),
-                        "Format \"%s\" cannot be embedded inside WPG.", clone_info->magick);
+                        "Format \"%s\" cannot be embedded inside WPG.", format);
     ThrowException(exception,CorruptImageError,UnableToReadImageHeader,image->filename);
-    goto FINISH_UNL;
+    return image;
   }
 
+  /*
+    Restore seek offset after reading header
+  */
+  if (SeekBlob(image,PS_Offset,SEEK_SET) != PS_Offset)
+    {
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "ExtractPostscript(): Failed to seek to PS_Offset=%"MAGICK_OFF_F"d",
+                              (magick_off_t) PS_Offset);
+      ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
+      return image;
+    }
+  /*
+    Allocate buffer if zero-copy read is not possible.
+  */
+  if (GetBlobStreamData(image) == (unsigned char *) NULL)
+    {
+      ps_data_alloc=MagickAllocateResourceLimitedMemory(unsigned char *, PS_Size);
+      if (ps_data_alloc == (unsigned char *) NULL)
+        {
+          if (image->logging)
+            (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                  "ExtractPostscript(): Failed to allocate "
+                                  "%"MAGICK_SIZE_T_F"u bytes of memory",
+                                  (MAGICK_SIZE_T) PS_Size);
+          ThrowException(exception,ResourceLimitError,MemoryAllocationFailed,image->filename);
+          return image;
+        }
+    }
+  /*
+    Use a zero-copy read when possible to access data
+  */
+  ps_data=ps_data_alloc;
+  if (ReadBlobZC(image,PS_Size,&ps_data) != PS_Size)
+    {
+      MagickFreeResourceLimitedMemory(ps_data_alloc);
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "ExtractPostscript(): Failed to read %"MAGICK_SIZE_T_F"u bytes of data at"
+                              " offset=%"MAGICK_OFF_F"d",
+                              (MAGICK_SIZE_T) PS_Size, (magick_off_t) PS_Offset);
+      ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
+      return image; /* return (Image *) NULL; */
+    }
+  if (ps_data_alloc != ps_data)
+    {
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "ExtractPostscript(): Zero copy read.");
+    }
+#if 0
+  /*
+    Write in-memory blob to file for test purposes.
+  */
+  {
+    char file_name[MaxTextExtent];
+    FILE *file;
+
+    FormatString(file_name,"wpg-blob.%s",format);
+    if ((file=fopen(file_name,"w")) != (FILE *) NULL)
+      {
+        if (image->logging)
+          (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                "Writing %s...", file_name);
+        (void) fwrite(ps_data, 1, PS_Size, file);
+        (void) fclose(file);
+      }
+    SeekBlob(image,PS_Offset,SEEK_SET);
+  }
+#endif
+  /*
+    Read nested image from blob, forcing read as Postscript format
+  */
+  if ((clone_info=CloneImageInfo(image_info)) == NULL)
+    {
+      MagickFreeResourceLimitedMemory(ps_data_alloc);
+      return(image);
+    }
+  clone_info->blob=(void *) NULL;
+  /* clone_info->length=0; */
+  (void) strlcpy(clone_info->magick, format, sizeof(clone_info->magick));
+  (void) strlcpy(clone_info->filename, "", sizeof(clone_info->filename));
   (void) LogMagickEvent(CoderEvent,GetMagickModule(),
-                        "Reading embedded \"%s\" content...", clone_info->magick);
-
-  /* Read nested image, forcing read as Postscript format */
-  FormatString(clone_info->filename,"%s:%.1024s",clone_info->magick,postscript_file);
-  image2 = ReadImage(clone_info,exception);
-
+                        "Reading embedded \"%s\" content from blob...", clone_info->magick);
+  image2 = BlobToImage(clone_info, ps_data, PS_Size, &image->exception );
+  MagickFreeResourceLimitedMemory(ps_data_alloc);
   if (!image2)
-    goto FINISH_UNL;
+    {
+      goto FINISH_UNL;
+    }
   if(exception->severity>=ErrorException) /* When exception is raised, destroy image2 read. */
   {
     CloseBlob(image2);
@@ -907,12 +1011,66 @@ BAD_SEEK:
     image = image->next;                /* Rewind the cursor to the end. */
 
  FINISH_UNL:
-  (void) LiberateTemporaryFile(postscript_file);
- FINISH:
   DestroyImageInfo(clone_info);
   return(image);
 }
 
+#define LogHeaderWPG(_WPG_HEADER) \
+  (void)LogMagickEvent(CoderEvent,GetMagickModule(), \
+                       "WPG Header Id=%Xh:\n" \
+                       "    DataOffset=%u\n" \
+                       "    ProductType=%u\n" \
+                       "    FileType=%u\n" \
+                       "    Version=%u.%u\n" \
+                       "    EncryptKey=%u\n" \
+                       "    Reserved=%u", \
+                            (unsigned int)_WPG_HEADER.FileId, \
+                            (unsigned int)_WPG_HEADER.DataOffset, \
+                            (unsigned int)_WPG_HEADER.ProductType, \
+                            (unsigned int)_WPG_HEADER.FileType, \
+                            (unsigned int)_WPG_HEADER.MajorVersion, \
+                            (unsigned int)_WPG_HEADER.MinorVersion, \
+                            (unsigned int)_WPG_HEADER.EncryptKey, \
+                            (unsigned int)_WPG_HEADER.Reserved)
+
+#define LogWPGBitmapType1(_WPG_BITMAP_TYPE1) \
+  (void)LogMagickEvent(CoderEvent,GetMagickModule(), \
+                       "WPG Bitmap1 Header:\n" \
+                       "    Width=%u\n" \
+                       "    Heigth=%u\n" \
+                       "    Depth=%u\n" \
+                       "    HorzRes=%u\n" \
+                       "    VertRes=%u", \
+                            (unsigned int)_WPG_BITMAP_TYPE1.Width, \
+                            (unsigned int)_WPG_BITMAP_TYPE1.Heigth, \
+                            (unsigned int)_WPG_BITMAP_TYPE1.Depth, \
+                            (unsigned int)_WPG_BITMAP_TYPE1.HorzRes, \
+                            (unsigned int)_WPG_BITMAP_TYPE1.VertRes)
+
+#define LogWPGBitmapType2(_WPG_BITMAP_TYPE2) \
+  (void)LogMagickEvent(CoderEvent,GetMagickModule(), \
+                       "WPG Bitmap2 Header:\n" \
+                       "    RotAngle=%u\n" \
+                       "    LowLeftX=%u\n" \
+                       "    LowLeftY=%u\n" \
+                       "    UpRightX=%u\n" \
+                       "    UpRightY=%u\n" \
+                       "    Width=%u\n" \
+                       "    Heigth=%u\n" \
+                       "    Depth=%u\n" \
+                       "    HorzRes=%u\n" \
+                       "    VertRes=%u", \
+                            (unsigned int)_WPG_BITMAP_TYPE2.RotAngle, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.LowLeftX, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.LowLeftY, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.UpRightX, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.UpRightY, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.Width, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.Heigth, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.Depth, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.HorzRes, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.VertRes)
+
 
 /*
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -943,7 +1101,6 @@ BAD_SEEK:
 %
 %    o exception: return any errors or warnings in this structure.
 %
-%
 */
 static Image *ReadWPGImage(const ImageInfo *image_info,
   ExceptionInfo *exception)
@@ -1099,6 +1256,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   Header.MinorVersion=ReadBlobByte(image);
   Header.EncryptKey=ReadBlobLSBShort(image);
   Header.Reserved=ReadBlobLSBShort(image);
+  if(logging) LogHeaderWPG(Header);
 
   if (Header.FileId!=0x435057FF || (Header.ProductType>>8)!=0x16)
     ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
@@ -1106,17 +1264,18 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
     ThrowReaderException(CoderError,EncryptedWPGImageFileNotSupported,image);
 
   image->colors = 0;
+  image->storage_class = DirectClass;
   bpp=0;
 
   if (logging) (void)LogMagickEvent(CoderEvent,GetMagickModule(),
           "File type: %d", Header.FileType);
 
-        /* Determine file size. */
+  /* Determine file size. */
   filesize = GetBlobSize(image);              /* zero is returned if the size cannot be determined. */
   if(filesize>0 && BlobIsSeekable(image))
   {
     if(filesize > (magick_off_t)0xFFFFFFFF)
-        filesize = (magick_off_t)0xFFFFFFFF;  /* More than 4GiB are not supported in MAT! */
+        filesize = (magick_off_t)0xFFFFFFFF;  /* More than 4GiB are not supported in WPG! */
   }
   else
   {
@@ -1156,7 +1315,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
 
           if (logging) (void)LogMagickEvent(CoderEvent,GetMagickModule(),
             "Parsing object: %X", Rec.RecType);
-          //printf("\nParsing object: %u:%X", (unsigned)FilePos, Rec.RecType);
+          /* printf("\nParsing object: %u:%X", (unsigned)FilePos, Rec.RecType); */
 
           switch(Rec.RecType)
             {
@@ -1168,6 +1327,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               BitmapHeader1.Depth=ReadBlobLSBShort(image);
               BitmapHeader1.HorzRes=ReadBlobLSBShort(image);
               BitmapHeader1.VertRes=ReadBlobLSBShort(image);
+              if(logging) LogWPGBitmapType1(BitmapHeader1);
 
               if(BitmapHeader1.HorzRes && BitmapHeader1.VertRes)
                 {
@@ -1178,7 +1338,8 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               image->columns=BitmapHeader1.Width;
               image->rows=BitmapHeader1.Heigth;
               bpp=BitmapHeader1.Depth;
-
+				// Whole palette is useless for bilevel image.
+              if(bpp==1) image->storage_class=DirectClass;
               goto UnpackRaster;
 
             case 0x0E:  /*Color palette */
@@ -1203,7 +1364,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               if(Rec.RecordLength > 8)
                 image=ExtractPostscript(image,image_info,
                                         TellBlob(image)+8,   /* skip PS header in the wpg */
-                                        (long) Rec.RecordLength-8,exception);
+                                        (size_t) (Rec.RecordLength-8),exception);
               break;
 
             case 0x14:  /* bitmap type 2 */
@@ -1219,6 +1380,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               BitmapHeader2.Depth=ReadBlobLSBShort(image);
               BitmapHeader2.HorzRes=ReadBlobLSBShort(image);
               BitmapHeader2.VertRes=ReadBlobLSBShort(image);
+              if(logging) LogWPGBitmapType2(BitmapHeader2);
 
               image->units=PixelsPerCentimeterResolution;
               image->page.width=(unsigned int)
@@ -1236,11 +1398,11 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               image->rows=BitmapHeader2.Heigth;
               bpp=BitmapHeader2.Depth;
 
-            UnpackRaster:
+UnpackRaster:
               if(bpp>24)
                 {ThrowReaderException(CoderError,ColorTypeNotSupported,image)}
 
-              if ((image->storage_class != PseudoClass) && (bpp != 24))
+              if ((image->storage_class != PseudoClass) && (bpp != 24) && bpp!=1)
                 {
                   image->colors=1 << bpp;
                   if (!AllocateImageColormap(image,image->colors))
@@ -1262,9 +1424,9 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               else
                 {
                   if (bpp < 24)
-                    if ( (image->colors < (1UL<<bpp)) && (bpp != 24) )
-                      MagickReallocMemory(PixelPacket *,image->colormap,
-                                          (size_t) (1U<<bpp)*sizeof(PixelPacket));
+                    if ( (image->colors != (1UL<<bpp)) && (bpp != 24) )
+                      if (!ReallocateImageColormap(image,1U<<bpp))
+                        goto NoMemory;
                 }
 
               if(bpp == 1)
@@ -1337,6 +1499,12 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                     }
                 }
 
+              StopTimer(&image->timer);
+
+              if (image_info->subrange != 0)
+                if (image->scene >= (image_info->subimage+image_info->subrange-1))
+                  goto Finish;
+
               /* Allocate next image structure. */
               AllocateNextImage(image_info,image);
               image->depth=8;
@@ -1351,7 +1519,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               if(Rec.RecordLength>0x3C)
                 image=ExtractPostscript(image,image_info,
                                         TellBlob(image)+0x3C,   /* skip PS l2 header in the wpg */
-                                        (long) Rec.RecordLength-0x3C,exception);
+                                        (size_t) (Rec.RecordLength-0x3C),exception);
               break;
             }
         }
@@ -1458,9 +1626,9 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               else
                 {
                   if(bpp < 24)
-                    if( image->colors<(1UL<<bpp) && bpp!=24 )
-                      MagickReallocMemory(PixelPacket *,image->colormap,
-                                          (size_t) (1U<<bpp)*sizeof(PixelPacket));
+                    if(image->colors!=(1UL<<bpp) && bpp!=24)
+                      if (!ReallocateImageColormap(image,1U<<bpp))
+                        goto NoMemory;
                 }
 
 
@@ -1469,26 +1637,26 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                 case 0:    /*Uncompressed raster*/
                   {
                     ldblk=(long) ((bpp*image->columns+7)/8);
-                    BImgBuff=MagickAllocateMemory(unsigned char *,(size_t) ldblk);
-                    if (BImgBuff == (unsigned char *) NULL)
+                    BImgBuff=MagickAllocateResourceLimitedMemory(unsigned char *,(size_t) ldblk);
+                    if(BImgBuff == (unsigned char *) NULL)
                       goto NoMemory;
 
                     for(i=0; i< (long) image->rows; i++)
                       {
                         if (ReadBlob(image,ldblk,(char *) BImgBuff) != (size_t) ldblk)
                           {
-                            MagickFreeMemory(BImgBuff);
+                            MagickFreeResourceLimitedMemory(BImgBuff);
                             ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,image->filename);
                             goto DecompressionFailed;
                           }
                         if(InsertRow(BImgBuff,i,image,bpp) == MagickFail)
                         {
-                          if(BImgBuff) MagickFreeMemory(BImgBuff);
+                          if(BImgBuff) MagickFreeResourceLimitedMemory(BImgBuff);
                           goto DecompressionFailed;
                         }
                       }
 
-                    if(BImgBuff) MagickFreeMemory(BImgBuff);
+                    if(BImgBuff) MagickFreeResourceLimitedMemory(BImgBuff);
                     break;
                   }
                 case 1:    /*RLE for WPG2 */
@@ -1534,6 +1702,11 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                      Tx(2,2)=1; */
                 }
 
+              StopTimer(&image->timer);
+
+              if (image_info->subrange != 0)
+                if (image->scene >= (image_info->subimage+image_info->subrange-1))
+                  goto Finish;
 
               /* Allocate next image structure. */
               AllocateNextImage(image_info,image);
@@ -1547,10 +1720,10 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
 
             case 0x12:  /* Postscript WPG2*/
               i=ReadBlobLSBShort(image);
-              if(Rec2.RecordLength > (unsigned int) i)
+              if(Rec2.RecordLength > ((unsigned long) i+2))
                 image=ExtractPostscript(image,image_info,
                                         TellBlob(image)+i,              /*skip PS header in the wpg2*/
-                                        (long) (Rec2.RecordLength-i-2),exception);
+                                        (size_t)Rec2.RecordLength-i-2,exception);
               break;
 
             case 0x1B:          /*bitmap rectangle*/
@@ -1641,6 +1814,7 @@ ModuleExport void RegisterWPGImage(void)
   entry->description="Word Perfect Graphics";
   entry->module="WPG";
   entry->seekable_stream=True;
+  entry->coder_class=UnstableCoderClass;
   (void) RegisterMagickInfo(entry);
 }
 
