@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003 - 2017 GraphicsMagick Group
+% Copyright (C) 2003 - 2022 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -189,6 +189,9 @@ MagickExport Image *ConstituteImage(const unsigned long width,
   size_t
     length;
 
+  unsigned int
+    storage_type_depth = QuantumDepth;
+
   /*
     Allocate image structure.
   */
@@ -205,6 +208,18 @@ MagickExport Image *ConstituteImage(const unsigned long width,
   image->columns=width;
   image->rows=height;
 
+  switch (type)
+    {
+    case CharPixel : storage_type_depth = sizeof(unsigned char)*8 ; break;
+    case ShortPixel : storage_type_depth = sizeof(unsigned short)*8 ; break;
+    case IntegerPixel : storage_type_depth = sizeof(unsigned short)*8 ; break;
+    case LongPixel : storage_type_depth = sizeof(unsigned long)*8 ; break;
+    case FloatPixel : storage_type_depth = sizeof(float)*8 ; break;
+    case DoublePixel : storage_type_depth = sizeof(double)*8 ; break;
+    }
+
+  image->depth = Min(storage_type_depth,QuantumDepth);
+
   /*
     Handle a few common special cases in order to improve performance.
   */
@@ -554,7 +569,8 @@ MagickExport Image *ConstituteImage(const unsigned long width,
                   }
                 }
             }
-          indexes++;
+          if (NULL != (IndexPacket *) indexes)
+            indexes++;
           q++;
         }
       if (!SyncImagePixels(image))
@@ -1122,7 +1138,7 @@ MagickFindRawImageMinMax(Image *image, EndianType endian,
 
   *min=0.0;
   *max=1.0;
-  status=MagickFail;
+  status=MagickPass;
 
   filepos = TellBlob(image);
 
@@ -1138,8 +1154,8 @@ MagickFindRawImageMinMax(Image *image, EndianType endian,
             read_func = ReadBlob;
 
             MagickFindMinMax(status,image,read_func,char,scanline_octets,
-                             scanline_buffer,min,max)
-              break;
+                             scanline_buffer,min,max);
+            break;
           }
         case ShortPixel:
           {
@@ -1196,7 +1212,8 @@ MagickFindRawImageMinMax(Image *image, EndianType endian,
           }
         }
 
-      (void) SeekBlob(image, filepos, SEEK_SET);
+      if (SeekBlob(image, filepos, SEEK_SET) != filepos)
+        status = MagickFail;
     }
 
   return status;
@@ -1428,17 +1445,17 @@ static void RemoveTemporaryInputFile(ImageInfo *image_info)
       (LocaleCompare(image_info->filename+filename_length-4,".mpc") == 0))
     {
       char remove_name[MaxTextExtent];
-      (void) strcpy(remove_name,image_info->filename);
+      (void) strlcpy(remove_name,image_info->filename,sizeof(remove_name));
       remove_name[filename_length-4]=0;
-      (void) strcat(remove_name,".cache");
+      (void) strlcat(remove_name,".cache",sizeof(remove_name));
       (void) printf("removing %s\n", remove_name);
       (void) remove(remove_name);
     }
   else if (LocaleCompare(image_info->magick,"mpc") == 0)
     {
       char remove_name[MaxTextExtent];
-      (void) strcpy(remove_name,image_info->filename);
-      (void) strcat(remove_name,".cache");
+      (void) strlcpy(remove_name,image_info->filename,sizeof(remove_name));
+      (void) strlcat(remove_name,".cache",sizeof(remove_name));
       (void) printf("removing %s\n", remove_name);
       (void) remove(remove_name);
     }
@@ -1586,7 +1603,14 @@ MagickExport Image *ReadImage(const ImageInfo *image_info,
               DestroyImage(image);
               return((Image *) NULL);
             }
-          (void) ImageToFile(image,clone_info->filename,exception);
+          if (ImageToFile(image,clone_info->filename,exception) == MagickFail)
+            {
+              LiberateTemporaryFile(clone_info->filename);
+              CloseBlob(image);
+              DestroyImageInfo(clone_info);
+              DestroyImage(image);
+              return(MagickFail);
+            }
           clone_info->temporary=True;
         }
       CloseBlob(image);
@@ -1979,7 +2003,7 @@ static Image *ReadImages(const ImageInfo *image_info,ExceptionInfo *exception)
   command=(char *) FileToBlob(image_info->filename+1,&length,exception);
   if (command == (char *) NULL)
     return((Image *) NULL);
-  Strip(command);
+  (void) MagickStripString(command);
   images=StringToArgv(command,&number_images);
   MagickFreeMemory(command);
   if (images == (char **) NULL)
@@ -2142,8 +2166,9 @@ WriteImage(const ImageInfo *image_info,Image *image)
   assert(image_info->filename != (char *) NULL);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
-  GetTimerInfo(&image->timer);
-  image->logging=IsEventLogging();
+  if (LocaleNCompare(image_info->magick,"INFO",sizeof("INFO")-1))
+      GetTimerInfo(&image->timer);
+  image->logging=IsEventLogged(CoderEvent);
   clone_info=CloneImageInfo(image_info);
   (void) strlcpy(clone_info->filename,image->filename,MaxTextExtent);
   (void) strlcpy(clone_info->magick,image->magick,MaxTextExtent);
@@ -2293,11 +2318,12 @@ WriteImage(const ImageInfo *image_info,Image *image)
       if ((magick_info == (MagickInfo *) NULL) ||
           (magick_info->encoder == NULL))
         {
+          ThrowException(&image->exception,MissingDelegateError,
+                         NoEncodeDelegateForThisImageFormat,
+                         clone_info->magick);
           DestroyImageInfo(clone_info);
-          ThrowBinaryException(MissingDelegateError,
-                               NoEncodeDelegateForThisImageFormat,
-                               image->filename)
-            }
+          return(MagickFail);
+        }
       if (!magick_info->thread_support)
         LockSemaphoreInfo(constitute_semaphore);
       status=(magick_info->encoder)(clone_info,image);
@@ -2377,7 +2403,7 @@ WriteImages(const ImageInfo *image_info,Image *image,
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
   assert(exception != (ExceptionInfo *) NULL);
-  image->logging=IsEventLogging();
+  image->logging=IsEventLogged(CoderEvent);
   clone_info=CloneImageInfo(image_info);
   if (clone_info)
     {
@@ -2465,7 +2491,7 @@ MagickExport MagickPassFail WriteImagesFile(const ImageInfo *image_info,Image *i
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
   assert(exception != (ExceptionInfo *) NULL);
-  image->logging=IsEventLogging();
+  image->logging=IsEventLogged(CoderEvent);
   clone_info=CloneImageInfo(image_info);
   if (clone_info)
     {
